"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.LambdaCircuitBreaker = void 0;
const opossum_1 = __importDefault(require("opossum"));
const serialize_error_1 = require("serialize-error");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
// executes `functionCall` with `args`
async function wrapperCircuitBreakerAction(functionCall, args) {
    return functionCall(...args);
}
class LambdaCircuitBreaker {
    constructor(circuitBreakerOptions, context, dependencyName, executeFunction, healthCheckFunction, healthCheckParams) {
        this.healthCheckParams = [];
        this.fallbackToRestartTimeoutMs = 180000; // 3 minutes by default
        this.circuitBreakerOpenCount = 0;
        this.context = context;
        this.dependencyName = dependencyName;
        this.executeFunction = executeFunction;
        this.healthCheckFunction = healthCheckFunction;
        this.healthCheckParams = healthCheckParams ?? [];
        this.fallbackToRestartTimeoutMs =
            circuitBreakerOptions.fallbackToRestartTimeoutMs ?? this.fallbackToRestartTimeoutMs;
        const lambdaErrorFilter = circuitBreakerOptions.errorFilter;
        this.circuitBreaker = new opossum_1.default(wrapperCircuitBreakerAction, {
            errorThresholdPercentage: circuitBreakerOptions.errorThresholdPercentage,
            resetTimeout: circuitBreakerOptions.resetTimeout,
            timeout: circuitBreakerOptions.timeout,
            rollingCountTimeout: circuitBreakerOptions.rollingCountTimeout,
            rollingCountBuckets: circuitBreakerOptions.rollingCountBuckets,
            errorFilter: (error) => {
                if (error.healthCheckFailed) {
                    // open the circuit breaker if health check fails with any error, else use lambdaErrorFilter
                    return false;
                }
                return lambdaErrorFilter ? lambdaErrorFilter(error) : false;
            },
        });
        this.circuitBreaker.on("open", () => this.openCallback());
        this.circuitBreaker.on("close", () => this.closeCallback());
        this.circuitBreaker.on("halfOpen", () => this.halfOpenCallback());
        this.circuitBreaker.on("failure", (err) => this.failureCallback(err)); // Emitted when the circuit breaker action fails
        this.circuitBreaker.on("reject", (err) => this.rejectCallback(err)); // Emitted when the circuit breaker is in open state (failing fast) and action is fired
    }
    async execute(params, doHealthCheck) {
        const functionToFire = doHealthCheck && this.healthCheckFunction
            ? this.healthCheckFunction
            : this.executeFunction;
        return this.circuitBreaker.fire(functionToFire, params); // calls wrapperCircuitBreakerAction with these params
    }
    getCurrentState() {
        return this.circuitBreaker.opened
            ? "opened"
            : this.circuitBreaker.halfOpen
                ? "halfOpen"
                : "closed";
    }
    shutdown() {
        this.circuitBreaker.shutdown();
    }
    openCallback() {
        // telemetry for circuit breaker open
        this.circuitBreakerOpenCount++;
        if (this.circuitBreakerMetric) {
            // opening the circuit agan after halfOpen state
            this.circuitBreakerMetric.setProperty("openCount", this.circuitBreakerOpenCount);
        }
        else {
            // opening the circuit first time, create new metric
            this.circuitBreakerMetric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.CircuitBreaker, {
                timestampOpened: new Date().toISOString(),
                dependencyName: this.dependencyName,
                error: (0, serialize_error_1.serializeError)(this.error),
                openCount: this.circuitBreakerOpenCount,
            });
            // setup the fallback to restart the service if the circuit breaker is not closed for more than fallbackToRestartTimeoutMs
            this.setupRestartFallback(this.error);
        }
        server_services_telemetry_1.Lumberjack.info("Circuit breaker opened", {
            metricId: this.circuitBreakerMetric.id,
            error: (0, serialize_error_1.serializeError)(this.error),
        });
    }
    closeCallback() {
        // resume lambda.
        this.context.resume();
        // telemetry for circuit breaker closed
        const metricProperties = {
            timestampClosed: new Date().toISOString(),
            openCount: this.circuitBreakerOpenCount,
            state: this.circuitBreaker.toJSON()?.state,
        };
        if (this.circuitBreakerMetric) {
            this.circuitBreakerMetric?.setProperties(metricProperties);
            this.circuitBreakerMetric?.success("Circuit breaker closed");
        }
        else {
            server_services_telemetry_1.Lumberjack.info("Circuit breaker closed", metricProperties);
        }
        // Reset the circuit breaker telemetry and fallback
        this.resetTelemetryAndFallback();
    }
    halfOpenCallback() {
        server_services_telemetry_1.Lumberjack.info("Circuit breaker halfOpen", {
            metricId: this.circuitBreakerMetric?.id,
        });
        // check the health of the dependency service, and let circuit breaker change its state accordingly
        this.execute(this.healthCheckParams, true).catch((error) => {
            server_services_telemetry_1.Lumberjack.error("Circuit breaker health check failed in halfOpen state, circuit will remain open.", { metricId: this.circuitBreakerMetric?.id }, error);
        });
    }
    failureCallback(error) {
        this.error = error;
        error.circuitBreakerOpen = true;
    }
    rejectCallback(error) {
        error.circuitBreakerOpen = true;
    }
    setupRestartFallback(initialError) {
        this.fallbackToRestartTimeout = setTimeout(() => {
            if (!this.circuitBreaker.closed) {
                this.circuitBreakerMetric?.setProperties({
                    openCount: this.circuitBreakerOpenCount,
                    timestampFallbackToRestart: new Date().toISOString(),
                    state: this.circuitBreaker.toJSON()?.state,
                    fallbackToRestartTimeoutMs: this.fallbackToRestartTimeoutMs,
                });
                this.circuitBreakerMetric?.error("Circuit breaker not closed for a long time, going to restart the service");
                this.context.error(initialError, {
                    restart: true,
                    errorLabel: "circuitBreaker:fallbackToRestartTimeout",
                });
            }
        }, this.fallbackToRestartTimeoutMs);
    }
    resetTelemetryAndFallback() {
        // clear the fallback to restart timeout
        if (this.fallbackToRestartTimeout !== undefined) {
            clearTimeout(this.fallbackToRestartTimeout);
            this.fallbackToRestartTimeout = undefined;
        }
        // reset the circuit breaker metric and count
        this.circuitBreakerMetric = undefined;
        this.circuitBreakerOpenCount = 0;
        // reset the error
        this.error = undefined;
    }
}
exports.LambdaCircuitBreaker = LambdaCircuitBreaker;
//# sourceMappingURL=circuitBreaker.js.map